@@ -29,10 +29,18 @@ module Decoder = struct
   external check_packet : Ogg.Stream.packet -> bool
     = "ocaml_flac_decoder_check_ogg"
 
+  external finalize_decoder_private_values : ogg Flac.Decoder.dec -> unit
+    = "ocaml_flac_finalize_ogg_decoder_private_values"
+
   external create :
     Ogg.Stream.packet -> Ogg.Stream.stream -> ogg Flac.Decoder.dec
     = "ocaml_flac_decoder_ogg_create"
 
+  let create p os =
+    let dec = create p os in
+    Gc.finalise finalize_decoder_private_values dec;
+    dec
+
   external update_ogg_stream : ogg Flac.Decoder.t -> Ogg.Stream.stream -> unit
     = "ocaml_flac_decoder_ogg_update_os"
 end
@@ -41,37 +49,41 @@ module Encoder = struct
   type ogg
   type enc
 
-  let callbacks : ogg Flac.Encoder.callbacks =
-    Obj.magic (Flac.Encoder.get_callbacks (fun _ -> raise Flac.Internal))
+  type t = {
+    encoder : ogg Flac.Encoder.t;
+    callbacks : ogg Flac.Encoder.callbacks;
+    first_pages : Ogg.Page.t list;
+  }
 
-  type init_c = Ogg.Stream.packet -> unit
+  external finalize_encoder_private_values : enc -> unit
+    = "ocaml_flac_finalize_ogg_encoder_private_values"
 
   external create :
     (string * string) array ->
     Flac.Encoder.params ->
-    Ogg.Stream.stream ->
-    init_c ->
+    (Ogg.Page.t -> unit) ->
+    nativeint ->
     enc = "ocaml_flac_encoder_ogg_create"
 
-  let create ?(comments = []) params os =
+  external set_write_cb : enc -> (Ogg.Page.t -> unit) -> unit
+    = "ocaml_flac_encoder_ogg_set_write_cb"
+
+  let create ?(comments = []) ~serialno params write_cb =
     if params.Flac.Encoder.channels <= 0 then raise Flac.Encoder.Invalid_data;
     let comments = Array.of_list comments in
-    let ret = Queue.create () in
-    let init_c p = Queue.push p ret in
-    let enc = create comments params os init_c in
-    let rec f acc =
-      try f (Queue.pop ret :: acc)
-      with Queue.Empty -> (
-        match List.rev acc with [] -> raise Flac.Internal | x :: l -> (x, l))
+    let first_pages = Atomic.make [] in
+    let write_first_page p =
+      Atomic.set first_pages (p :: Atomic.get first_pages)
     in
-    let p, l = f [] in
-    (Obj.magic (enc, params), p, l)
-
-  external finish : enc -> unit = "ocaml_flac_encoder_ogg_finish"
-
-  let finish e =
-    let e, _ = Obj.magic e in
-    finish e
+    let enc = create comments params write_first_page serialno in
+    Gc.finalise finalize_encoder_private_values enc;
+    set_write_cb enc write_cb;
+    {
+      encoder = Obj.magic (enc, params);
+      callbacks =
+        Obj.magic (Flac.Encoder.get_callbacks (fun _ -> raise Flac.Internal));
+      first_pages = List.rev (Atomic.get first_pages);
+    }
 end
 
 module Skeleton = struct
